from django.utils import six

from .models import Message, Dispatch
from .utils import is_iterable, get_registered_messenger_object, get_registered_message_type, import_project_sitemessage_modules
from .exceptions import UnknownMessengerError
from .messages import PlainTextMessage


# Trying import sitemessage settings files from project apps.
import_project_sitemessage_modules()


def recipients(messenger, addresses):
    """Structures recipients data.

    :param str, MessageBase messenger: MessengerBase heir
    :param list addresses: recipients addresses or Django User model heir instances (NOTE: if supported by a messenger)
    :return: list of Recipient
    :rtype: list
    """
    if isinstance(messenger, six.string_types):
        messenger = get_registered_messenger_object(messenger)
    return messenger._structure_recipients_data(addresses)


def schedule_messages(messages, recipients=None, sender=None, priority=None):
    """Sends a message(s).

    :param MessageBase, str, list messages: str or MessageBase heir or list - use str to create PlainTextMessage.
    :param list recipients: recipients addresses or Django User model heir instances
    :param User sender: User model heir instance
    :param int priority: number describing message priority. If set overrides priority provided with message type.
    :return: list of tuples - (message_model, dispatches_models)
    :rtype: list
    """
    if not is_iterable(messages):
        messages = (messages,)

    results = []
    for message in messages:
        if isinstance(message, six.string_types):
            message = PlainTextMessage(message)

        resulting_priority = message.priority
        if priority is not None:
            resulting_priority = priority
        results.append(message.schedule(sender=sender, recipients=recipients, priority=resulting_priority))

    return results


def send_scheduled_messages(priority=None, ignore_unknown_messengers=False, ignore_unknown_message_types=False):
    """Sends scheduled messages.

    :param int, None priority: number to limit sending message by this priority.
    :param bool ignore_unknown_messengers: to silence UnknownMessengerError
    :param bool ignore_unknown_message_types: to silence UnknownMessageTypeError
    :raises UnknownMessengerError:
    :raises UnknownMessageTypeError:
    """
    dispatches_by_messengers = Dispatch.group_by_messengers(Dispatch.get_unsent(priority=priority))

    for messenger_id, messages in dispatches_by_messengers.items():
        try:
            messenger_obj = get_registered_messenger_object(messenger_id)
            messenger_obj._process_messages(messages, ignore_unknown_message_types=ignore_unknown_message_types)
        except UnknownMessengerError:
            if ignore_unknown_messengers:
                continue
            raise


def prepare_dispatches():
    """Created dispatches for messages without them.

    It requires from a message type to implement `calculate_recipients()` method.

    :return: list of Dispatch
    :rtype: list
    """
    dispatches = []
    undispatched = Message.get_undispatched()
    for message_model in undispatched:
        message_cls = get_registered_message_type(message_model.cls)
        dispatches.extend(message_cls.prepare_dispatches(message_model))
    return dispatches



try:
    from django.contrib.auth import get_user_model
    USER_MODEL = get_user_model()
except ImportError:
    # Django 1.4 fallback.
    from django.contrib.auth.models import User as USER_MODEL
from django.conf import settings
from django.utils.translation import ugettext_lazy as _
from django.utils.html import strip_tags

from .utils import MessengerBase
from .exceptions import MessengerWarmupException


class SMTPMessenger(MessengerBase):
    """Implements SMTP message delivery using Python builtin smtplib module."""

    alias = 'smtp'
    smtp = None
    _session_started = False

    def __init__(self, from_email=None, login=None, password=None, host=None, port=None, use_tls=None, use_ssl=None, debug=False):
        """Configures messenger.

        :param str from_email: e-mail address to send messages from
        :param str login: login to log into SMTP server
        :param str password: password to log into SMTP server
        :param str host: string - SMTP server host
        :param int port:  SMTP server port
        :param bool use_tls: whether to use TLS
        :param bool use_ssl: whether to use SSL
        :param bool debug: whether to switch smtplib into debug mode.
        """
        import smtplib

        from email.mime.text import MIMEText
        from email.mime.multipart import MIMEMultipart

        self.debug = debug
        self.lib = smtplib
        self.mime_text = MIMEText
        self.mime_multipart = MIMEMultipart

        self.from_email = from_email or getattr(settings, 'SERVER_EMAIL')
        self.login = login or getattr(settings, 'EMAIL_HOST_USER')
        self.password = password or getattr(settings, 'EMAIL_HOST_PASSWORD')
        self.host = host or getattr(settings, 'EMAIL_HOST')
        self.port = port or getattr(settings, 'EMAIL_PORT')
        self.use_tls = use_tls or getattr(settings, 'EMAIL_USE_TLS')
        self.use_ssl = use_ssl or getattr(settings, 'EMAIL_USE_SSL', False)  # False as default to support Django < 1.7

    @classmethod
    def get_address(cls, recipient):
        if hasattr(recipient, 'email'):
            recipient = recipient.email
        return recipient

    def _test_message(self, to, text):
        return self._send_message(self._build_message(to, text, mtype='html'))

    def before_send(self):
        try:
            self.smtp = self.lib.SMTP(self.host, self.port)
            self.smtp.set_debuglevel(self.debug)

            if self.use_tls:
                self.smtp.ehlo()
                if self.smtp.has_extn('STARTTLS'):
                    self.smtp.starttls()
                    self.smtp.ehlo()  # This time over TLS.

            if self.login:
                self.smtp.login(self.login, self.password)

            self._session_started = True
        except self.lib.SMTPException as e:
            raise MessengerWarmupException('SMTP Error: %s' % e)

    def after_send(self):
        self.smtp.quit()

    def _build_message(self, to, text, subject=None, mtype=None):
        """Constructs a MIME message from message and dispatch models."""
        # TODO Maybe file attachments handling through `files` message_model context var.

        if subject is None:
            subject = u'%s' % _('No Subject')

        if mtype == 'html':
            msg = self.mime_multipart()
            text_part = self.mime_multipart('alternative')
            text_part.attach(self.mime_text(strip_tags(text), _charset='utf-8'))
            text_part.attach(self.mime_text(text, 'html', _charset='utf-8'))
            msg.attach(text_part)
        else:
            msg = self.mime_text(text, _charset='utf-8')

        msg['From'] = self.from_email
        msg['To'] = to
        msg['Subject'] = subject

        return msg

    def _send_message(self, msg):
        return self.smtp.sendmail(msg['From'], msg['To'], msg.as_string())

    def send(self, message_cls, message_model, dispatch_models):
        if self._session_started:
            for dispatch_model in dispatch_models:

                msg = self._build_message(
                    dispatch_model.address,
                    dispatch_model.message_cache,
                    message_model.context.get('subject'),
                    message_model.context.get('type'),
                )

                try:
                    refused = self._send_message(msg)
                    if refused:
                        self.mark_failed(dispatch_model, '`%s` address is rejected by server' % msg['To'])
                        continue
                    self.mark_sent(dispatch_model)
                except Exception as e:
                    self.mark_error(dispatch_model, e, message_cls)


class XMPPSleekMessenger(MessengerBase):
    """Implements XMPP message delivery using `sleekxmpp` module."""

    alias = 'xmppsleek'
    xmpp = None
    _session_started = False

    def __init__(self, from_jid, password, host='localhost', port=5222, use_tls=True, use_ssl=False):
        """Configures messenger.

        :param str from_jid: Jabber ID to send messages from
        :param str password: password to log into XMPP server
        :param str host: XMPP server host
        :param int port: XMPP server port
        :param bool use_tls: whether to use TLS
        :param bool use_ssl: whether to use SSL
        :return:
        """
        import sleekxmpp

        self.lib = sleekxmpp

        self.from_jid = from_jid
        self.password = password
        self.host = host
        self.port = port
        self.use_tls = use_tls
        self.use_ssl = use_ssl

    def _test_message(self, to, text):
        return self._send_message(to, text)

    def _send_message(self, to, text):
        return self.xmpp.send_message(mfrom=self.from_jid, mto=to, mbody=text, mtype='chat')

    def before_send(self):
        def on_session_start(event):
            try:
                self.xmpp.send_presence()
                self.xmpp.get_roster()
                self._session_started = True
            except self.lib.exceptions.XMPPError as e:
                raise MessengerWarmupException('XMPP Error: %s' % e)

        self.xmpp = self.lib.ClientXMPP(self.from_jid, self.password)
        self.xmpp.add_event_handler('session_start', on_session_start)

        result = self.xmpp.connect(address=(self.host, self.port), reattempt=False, use_tls=self.use_tls, use_ssl=self.use_ssl)
        if result:
            self.xmpp.process(block=False)

    def after_send(self):
        if self._session_started:
            self.xmpp.disconnect(wait=True)  # Wait for a send queue.
            self._session_started = False

    def send(self, message_cls, message_model, dispatch_models):
        if self._session_started:
            for dispatch_model in dispatch_models:
                try:
                    self._send_message(dispatch_model.address, dispatch_model.message_cache)
                    self.mark_sent(dispatch_model)
                except Exception as e:
                    self.mark_error(dispatch_model, e, message_cls)



from .messages import EmailHtmlMessage, EmailTextMessage
from .toolbox import schedule_messages, recipients


def schedule_email(message, to, subject=None, sender=None, priority=None):
    """Schedules an email message for delivery.

    :param dict, str message: str or dict: use str for simple text email; dict - to compile email from a template (default: `sitemessage/email_html_smtp.html`).
    :param list to: recipients addresses or Django User model heir instances
    :param str subject: email subject
    :param User sender: User model heir instance
    :param int priority: number describing message priority. If set overrides priority provided with message type.
    """
    if isinstance(message, dict):
        message_cls = EmailHtmlMessage
    else:
        message_cls = EmailTextMessage
    schedule_messages(message_cls(subject, message), recipients('smtp', to), sender=sender, priority=priority)


def schedule_jabber_message(message, to, sender=None, priority=None):
    """Schedules an email message for delivery.

    :param str message: text to send.
    :param list to: recipients addresses or Django User model heir instances
    :param User sender: User model heir instance
    :param int priority: number describing message priority. If set overrides priority provided with message type.
    """
    schedule_messages(message, recipients('xmppsleek', to), sender=sender, priority=priority)



from django.utils import unittest
from django.contrib.auth.models import User
from django.template.base import TemplateDoesNotExist
from django.db.utils import IntegrityError

from .messages import PlainTextMessage
from .models import Message, Dispatch
from .toolbox import schedule_messages, recipients, send_scheduled_messages, prepare_dispatches
from .utils import MessageBase, MessengerBase, Recipient, register_messenger_objects, \
    register_message_types, get_registered_messenger_objects, get_registered_messenger_object, \
    get_registered_message_types
from .exceptions import MessengerWarmupException


# TODO More tests, please %)

WONDERLAND_DOMAIN = '@wonderland'


class TestMessenger(MessengerBase):

    alias = 'test_messenger'
    last_send = None

    def __init__(self, login, password):
        self.login = login
        self.password = password

    @classmethod
    def get_address(cls, recipient):
        if isinstance(recipient, User):
            recipient = 'user_%s' % recipient.username

        return '%s%s' % (recipient, WONDERLAND_DOMAIN)

    def send(self, message_cls, message_model, dispatch_models):
        self.last_send = {
            'message_cls': message_cls,
            'dispatch_models': dispatch_models,
            'message_model': message_model,
        }


class BuggyMessenger(MessengerBase):

    alias = 'buggy'

    def send(self, message_cls, message_model, dispatch_models):
        raise Exception('Damn it.')


class TestMessage(MessageBase):

    alias = 'test_message'
    template_ext = 'html'


class TestMessagePlain(PlainTextMessage):

    alias = 'testplain'
    priority = 10

    @classmethod
    def calculate_recipients(cls, message):
        return recipients('test_messenger', ('fred', 'colon'))


class TestMessagePlainDynamic(PlainTextMessage):

    alias = 'testplain_dyn'
    has_dynamic_context = True

    @classmethod
    def compile(cls, message, messenger, dispatch=None):
        return '%s -- %s' % (message.context[MessageBase.SIMPLE_TEXT_ID], dispatch.address)


register_messenger_objects(TestMessenger('mylogin', 'mypassword'), BuggyMessenger())
register_message_types(PlainTextMessage, TestMessage, TestMessagePlain, TestMessagePlainDynamic)


class UtilityTest(unittest.TestCase):

    def setUp(self):
        Message.objects.all().delete()

    def test_register_messengers(self):
        messenger = type('MyMessenger', (MessengerBase,), {})
        register_messenger_objects(messenger)
        self.assertIn(messenger.get_alias(), get_registered_messenger_objects())

    def test_register_message_types(self):
        message = type('MyMessage', (MessageBase,), {})
        register_message_types(message)
        self.assertIn(message.get_alias(), get_registered_message_types())

    def test_recipients(self):
        user = User(username='myuser')
        to = ('gogi', 'givi', user)

        r1 = recipients('test_messenger', to)

        self.assertEqual(len(r1), len(to))
        self.assertEqual(r1[0].address, 'gogi%s' % WONDERLAND_DOMAIN)
        self.assertEqual(r1[0].messenger, 'test_messenger')
        self.assertEqual(r1[1].address, 'givi%s' % WONDERLAND_DOMAIN)
        self.assertEqual(r1[1].messenger, 'test_messenger')
        self.assertEqual(r1[2].address, 'user_myuser%s' % WONDERLAND_DOMAIN)
        self.assertEqual(r1[2].messenger, 'test_messenger')

    def test_prepare_undispatched(self):
        m, d = Message.create('testplain', {MessageBase.SIMPLE_TEXT_ID: 'abc'})
        dispatches = prepare_dispatches()
        self.assertEqual(len(dispatches), 2)
        self.assertEqual(dispatches[0].address, 'fred%s' % WONDERLAND_DOMAIN)
        self.assertEqual(dispatches[1].address, 'colon%s' % WONDERLAND_DOMAIN)

    def test_send_scheduled_messages(self):
        # This one won't count, as won't fit into message priority filter.
        schedule_messages(TestMessagePlainDynamic('my_dyn_msg'), recipients('test_messenger', ('three', 'four')))

        msgr = get_registered_messenger_object('test_messenger')
        msg = TestMessagePlain('my_message')
        schedule_messages(msg, recipients(msgr, ('one', 'two')))
        send_scheduled_messages(priority=TestMessagePlain.priority)
        self.assertEqual(len(msgr.last_send['dispatch_models']), 2)
        self.assertEqual(msgr.last_send['message_model'].cls, 'testplain')
        self.assertEqual(msgr.last_send['message_cls'], TestMessagePlain)
        self.assertEqual(msgr.last_send['dispatch_models'][0].message_cache, 'my_message')
        self.assertEqual(msgr.last_send['dispatch_models'][1].message_cache, 'my_message')

    def test_send_scheduled_messages_dynamic_context(self):
        msgr = get_registered_messenger_object('test_messenger')
        msg_dyn = TestMessagePlainDynamic('my_dyn_msg')
        schedule_messages(msg_dyn, recipients(msgr, ('three', 'four')))
        send_scheduled_messages()
        self.assertEqual(len(msgr.last_send['dispatch_models']), 2)
        self.assertEqual(msgr.last_send['message_model'].cls, 'testplain_dyn')
        self.assertEqual(msgr.last_send['message_cls'], TestMessagePlainDynamic)
        self.assertEqual(msgr.last_send['dispatch_models'][0].message_cache, 'my_dyn_msg -- three%s' % WONDERLAND_DOMAIN)
        self.assertEqual(msgr.last_send['dispatch_models'][1].message_cache, 'my_dyn_msg -- four%s' % WONDERLAND_DOMAIN)

    def test_schedule_message(self):
        msg = TestMessagePlain('schedule_func')
        model, _ = schedule_messages(msg)[0]
        self.assertEqual(model.cls, msg.get_alias())
        self.assertEqual(model.context, msg.get_context())
        self.assertEqual(model.priority, TestMessagePlain.priority)
        self.assertFalse(model.dispatches_ready)

        msg = TestMessagePlain('schedule_func')
        model, _ = schedule_messages(msg, priority=33)[0]
        self.assertEqual(model.cls, msg.get_alias())
        self.assertEqual(model.context, msg.get_context())
        self.assertEqual(model.priority, 33)
        self.assertFalse(model.dispatches_ready)

        user = User()

        model, dispatch_models = schedule_messages('simple message', recipients('test_messenger', ('gogi', 'givi')), sender=user)[0]

        self.assertEqual(model.cls, 'plain')
        self.assertEqual(model.context, {'use_tpl': False, MessageBase.SIMPLE_TEXT_ID: 'simple message', 'tpl': None})
        self.assertEqual(model.sender, user)
        self.assertTrue(model.dispatches_ready)

        self.assertEqual(len(dispatch_models), 2)
        self.assertEqual(dispatch_models[0].address, 'gogi%s' % WONDERLAND_DOMAIN)
        self.assertEqual(dispatch_models[0].messenger, 'test_messenger')


class DispatchModelTest(unittest.TestCase):

    def test_create(self):

        message = Message(cls='test_message')
        message.save()

        user = User(username='u')

        recipients_ = recipients('test_messenger', [user])
        recipients_ += recipients('buggy', 'idle')

        dispatches = Dispatch.create(message, recipients_)
        self.assertTrue(isinstance(dispatches[0], Dispatch))
        self.assertTrue(isinstance(dispatches[1], Dispatch))
        self.assertEqual(dispatches[0].message_id, message.id)
        self.assertEqual(dispatches[1].message_id, message.id)
        self.assertEqual(dispatches[0].messenger, 'test_messenger')
        self.assertEqual(dispatches[1].messenger, 'buggy')


class MessageModelTest(unittest.TestCase):

    def test_create(self):
        user = User(username='u')

        m, _ = Message.create('some', {'abc': 'abc'}, sender=user, priority=22)
        self.assertEqual(m.cls, 'some')
        self.assertEqual(m.context, {'abc': 'abc'})
        self.assertEqual(m.sender, user)
        self.assertEqual(m.priority, 22)
        self.assertFalse(m.dispatches_ready)

        ctx = {'a': 'a', 'b': 'b'}
        m, _ = Message.create('some2', ctx)
        self.assertEqual(m.cls, 'some2')
        self.assertEqual(m.context, ctx)
        self.assertIsNone(m.sender)
        self.assertFalse(m.dispatches_ready)

    def test_deserialize_context(self):
        m = Message(cls='some_cls', context={'a': 'a', 'b': 'b', 'c': 'c'})
        m.save()

        m2 = Message.objects.get(pk=m.pk)
        self.assertEquals(m2.context, {'a': 'a', 'b': 'b', 'c': 'c'})


class MessengerTest(unittest.TestCase):

    def test_init_params(self):
        messengers = get_registered_messenger_objects()
        my = messengers['test_messenger']
        self.assertEqual(my.login, 'mylogin')
        self.assertEqual(my.password, 'mypassword')

    def test_alias(self):
        messenger = type('MyMessenger', (MessengerBase,), {'alias': 'myalias'})
        self.assertEqual(messenger.get_alias(), 'myalias')

        messenger = type('MyMessenger', (MessengerBase,), {})
        self.assertEqual(messenger.get_alias(), 'MyMessenger')

    def test_get_recipients_data(self):
        user = User(username='myuser')
        to = ('gogi', 'givi', user)

        r1 = TestMessenger._structure_recipients_data(to)

        self.assertEqual(len(r1), len(to))
        self.assertEqual(r1[0].address, 'gogi%s' % WONDERLAND_DOMAIN)
        self.assertEqual(r1[0].messenger, 'test_messenger')
        self.assertEqual(r1[1].address, 'givi%s' % WONDERLAND_DOMAIN)
        self.assertEqual(r1[1].messenger, 'test_messenger')
        self.assertEqual(r1[2].address, 'user_myuser%s' % WONDERLAND_DOMAIN)
        self.assertEqual(r1[2].messenger, 'test_messenger')

    def test_send(self):
        m = TestMessenger('l', 'p')
        m.send('message_cls', 'message_model', 'dispatch_models')

        self.assertEqual(m.last_send['message_cls'], 'message_cls')
        self.assertEqual(m.last_send['message_model'], 'message_model')
        self.assertEqual(m.last_send['dispatch_models'], 'dispatch_models')

        m = BuggyMessenger()
        recipiets_ = recipients('test_messenger', ['a', 'b', 'c', 'd'])
        self.assertRaises(Exception, m.send, 'a buggy message', recipiets_)


class MessageTest(unittest.TestCase):

    def test_alias(self):
        message = type('MyMessage', (MessageBase,), {'alias': 'myalias'})
        self.assertEqual(message.get_alias(), 'myalias')

        message = type('MyMessage', (MessengerBase,), {})
        self.assertEqual(message.get_alias(), 'MyMessage')

    def test_context(self):
        msg = TestMessage({'title': 'My message!', 'name': 'idle'})
        self.assertEqual(msg.context, {'name': 'idle', 'title': 'My message!', 'tpl': None, 'use_tpl': True})

    def test_get_template(self):
        self.assertEqual(TestMessage.get_template(TestMessage(), TestMessenger('a', 'b')), 'sitemessage/test_message_test_messenger.html')

    def test_compile_string(self):
        msg = TestMessagePlain('simple')
        self.assertEqual(TestMessagePlain.compile(msg, TestMessenger('a', 'b')), 'simple')

    def test_compile_dict(self):
        msg = TestMessage({'title': 'My message!', 'name': 'idle'})
        self.assertRaises(TemplateDoesNotExist, TestMessage.compile, msg, TestMessenger('a', 'b'))

    def test_schedule(self):
        msg = TestMessagePlain('schedule1')
        model, _ = msg.schedule()
        self.assertEqual(model.cls, msg.get_alias())
        self.assertEqual(model.context, msg.get_context())
        self.assertIsNone(model.sender)

        user = User()
        msg = TestMessagePlain('schedule2')
        model, _ = msg.schedule(sender=user)
        self.assertEqual(model.sender, user)



import json

from datetime import datetime

from django.core import exceptions
from django.db import models
from django.utils.six import with_metaclass
from django.utils.translation import ugettext_lazy as _
from django.utils.encoding import python_2_unicode_compatible
from django.conf import settings


USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')


# This allows South to handle our custom 'CharFieldNullable' field.
if 'south' in settings.INSTALLED_APPS:
    from south.modelsinspector import add_introspection_rules
    add_introspection_rules([], ['^sitemessage\.models\.ContextField'])


class ContextField(with_metaclass(models.SubfieldBase, models.TextField)):

    def to_python(self, value):
        if not value:
            return {}

        if isinstance(value, dict):
            return value

        try:
            return json.loads(value)
        except ValueError:
            raise exceptions.ValidationError(_('Value `%r` is not a valid context.') % value, code='invalid_context', params={'value': value},)

    def get_prep_value(self, value):
        return json.dumps(value)


@python_2_unicode_compatible
class Message(models.Model):

    time_created = models.DateTimeField(_('Time created'), auto_now_add=True, editable=False)
    sender = models.ForeignKey(USER_MODEL, verbose_name=_('Sender'), null=True, blank=True)
    cls = models.CharField(_('Message class'), max_length=250, help_text=_('Message logic class identifier.'), db_index=True)
    context = ContextField(_('Message context'))
    priority = models.PositiveIntegerField(_('Priority'), help_text=_('Number describing message sending priority. Messages with different priorities can be sent with different periodicity.'), default=0, db_index=True)
    dispatches_ready = models.BooleanField(_('Dispatches ready'), db_index=True, default=False, help_text=_('Indicates whether dispatches for this message are already formed and ready to delivery.'))

    @classmethod
    def get_undispatched(cls):
        return cls.objects.filter(dispatches_ready=False).all()

    @classmethod
    def create(cls, message_class, context, recipients=None, sender=None, priority=None):
        dispatches_ready = False
        if recipients is not None:
            dispatches_ready = True

        msg_kwargs = {
            'cls': message_class,
            'context': context,
            'sender': sender,
            'dispatches_ready': dispatches_ready
        }

        if priority is not None:
            msg_kwargs['priority'] = priority

        message_model = cls(**msg_kwargs)
        message_model.save()
        dispatch_models = Dispatch.create(message_model, recipients)
        return message_model, dispatch_models

    class Meta:
        verbose_name = _('Message')
        verbose_name_plural = _('Messages')

    def __str__(self):
        return self.cls


@python_2_unicode_compatible
class Dispatch(models.Model):

    DISPATCH_STATUS_PENDING = 1
    DISPATCH_STATUS_SENT = 2
    DISPATCH_STATUS_ERROR = 3
    DISPATCH_STATUS_FAILED = 4

    DISPATCH_STATUSES = (
        (DISPATCH_STATUS_PENDING, _('Pending')),
        (DISPATCH_STATUS_SENT, _('Sent')),
        (DISPATCH_STATUS_ERROR, _('Error')),
        (DISPATCH_STATUS_FAILED, _('Failed')),
    )

    READ_STATUS_UNDREAD = 0
    READ_STATUS_READ = 1

    READ_STATUSES = (
        (READ_STATUS_UNDREAD, _('Unread')),
        (READ_STATUS_READ, _('Read')),
    )

    error_log = None

    time_created = models.DateTimeField(_('Time created'), auto_now_add=True, editable=False)
    time_dispatched = models.DateTimeField(_('Time dispatched'), help_text=_('Time of the last delivery attempt.'), editable=False, null=True, blank=True)
    message = models.ForeignKey(Message, verbose_name=_('Message'))
    messenger = models.CharField(_('Messenger'), max_length=250, help_text=_('Messenger class identifier.'), db_index=True)
    recipient = models.ForeignKey(USER_MODEL, verbose_name=_('Recipient'), null=True, blank=True)
    address = models.CharField(_('Address'), max_length=250, help_text=_('Recipient address.'))
    retry_count = models.PositiveIntegerField(_('Retry count'), default=0, help_text=_('A number of delivery retries has already been made.'))
    message_cache = models.TextField(_('Message cache'), null=True, editable=False)
    dispatch_status = models.PositiveIntegerField(_('Dispatch status'), choices=DISPATCH_STATUSES, default=DISPATCH_STATUS_PENDING)
    read_status = models.PositiveIntegerField(_('Read status'), choices=READ_STATUSES, default=READ_STATUS_UNDREAD)

    @classmethod
    def log_dispatches_errors(cls, dispatches):
        entries = []
        for dispatch in dispatches:
            # Saving message cache for further usage.
            dispatch.save()
            entries = cls(dispatch=dispatch, error_log=dispatch.error_log)
        cls.objects.bulk_create(entries)

    @classmethod
    def set_dispatches_statuses(cls, **kwargs):
        filter_kwargs_map = {
            'sent': cls.DISPATCH_STATUS_SENT,
            'error': cls.DISPATCH_STATUS_ERROR,
            'failed': cls.DISPATCH_STATUS_FAILED,
            'pending': cls.DISPATCH_STATUS_PENDING,
        }
        for kwarg_name, status in filter_kwargs_map.items():
            if kwargs[kwarg_name]:
                update_kwargs = {
                    'time_dispatched': datetime.now(),
                    'dispatch_status': status,
                    'retry_count': models.F('retry_count') + 1
                }
                cls.objects.filter(id__in=[d.id for d in kwargs[kwarg_name]]).update(**update_kwargs)

    @staticmethod
    def group_by_messengers(dispatches):
        by_messengers = {}
        for dispatch in dispatches:
            if dispatch.messenger not in by_messengers:
                by_messengers[dispatch.messenger] = {}

            if dispatch.message.id not in by_messengers[dispatch.messenger]:
                by_messengers[dispatch.messenger][dispatch.message.id] = (dispatch.message, [])

            by_messengers[dispatch.messenger][dispatch.message.id][1].append(dispatch)
        return by_messengers

    @classmethod
    def get_unsent(cls, priority=None):
        filter_kwargs = {
            'dispatch_status__in': (cls.DISPATCH_STATUS_PENDING, cls.DISPATCH_STATUS_ERROR)
        }
        if priority is not None:
            filter_kwargs['message__priority'] = priority
        return cls.objects.select_related('message').filter(**filter_kwargs).order_by('-message__time_created').all()

    @classmethod
    def get_unread(cls):
        return cls.objects.filter(read_status=cls.READ_STATUS_UNDREAD).select_related('message').all()

    @classmethod
    def create(cls, message_model, recipients):
        objects = []
        if recipients:
            if not hasattr(recipients, '__iter__'):
                recipients = (recipients,)

            for r in recipients:
                objects.append(cls(message=message_model, messenger=r.messenger, recipient=r.user, address=r.address))

            if objects:
                cls.objects.bulk_create(objects)

            if not message_model.dispatches_ready:
                message_model.dispatches_ready = True
                message_model.save()

        return objects

    class Meta:
        verbose_name = _('Dispatch')
        verbose_name_plural = _('Dispatches')

    def __str__(self):
        return '%s [%s]' % (self.address, self.messenger)


@python_2_unicode_compatible
class DispatchError(models.Model):

    time_created = models.DateTimeField(_('Time created'), auto_now_add=True, editable=False)
    dispatch = models.ForeignKey(Dispatch, verbose_name=_('Dispatch'))
    error_log = models.TextField(_('Text'))

    class Meta:
        verbose_name = _('Dispatch error')
        verbose_name_plural = _('Dispatch errors')

    def __str__(self):
        return 'Dispatch ID %s error entry' % self.dispatch_id



class SiteMessageError(Exception):
    """Base class for sitemessage errors."""


class UnknownMessageTypeError(SiteMessageError):
    """This error is raised when there's a try to access an unknown message type."""


class UnknownMessengerError(SiteMessageError):
    """This error is raised when there's a try to access an unknown messenger."""


class MessengerWarmupException(SiteMessageError):
    """This exception represents a delivery error due to a messenger warm up process failure."""



from collections import namedtuple

try:
    from django.contrib.auth import get_user_model
    USER_MODEL = get_user_model()
except ImportError:
    # Django 1.4 fallback.
    from django.contrib.auth.models import User as USER_MODEL
from django.utils.importlib import import_module
from django.utils.module_loading import module_has_submodule
from django.template.loader import render_to_string
from django.utils import six

from sitemessage import settings
from .models import Message, Dispatch
from .exceptions import UnknownMessageTypeError, UnknownMessengerError


_MESSENGERS_REGISTRY = {}
_MESSAGES_REGISTRY = {}


def register_messenger_objects(*messengers):
    """Registers (configures) messengers.

    :param MessageBase messengers: MessengerBase heirs instances.
    """
    global _MESSENGERS_REGISTRY

    for messenger in messengers:
        _MESSENGERS_REGISTRY[messenger.get_alias()] = messenger


def get_registered_messenger_objects():
    """Returns registered (configured) messengers dict
    indexed by messenger aliases.

    :return: dict
    :rtype: dict
    """
    return _MESSENGERS_REGISTRY


def get_registered_messenger_object(messenger):
    """Returns registered (configured) messenger by alias,

    :param str messenger: messenger alias
    :return: MessengerBase heirs instances.
    :rtype: MessengerBase
    """
    try:
        return _MESSENGERS_REGISTRY[messenger]
    except KeyError:
        raise UnknownMessengerError('`%s` messenger is not registered' % messenger)


def register_message_types(*message_types):
    """Registers message types (classes).

    :param MessageBase message_types: MessageBase heir classes.
    """
    global _MESSAGES_REGISTRY

    for message in message_types:
        _MESSAGES_REGISTRY[message.get_alias()] = message


def get_registered_message_types():
    """Returns registered message types dict indexed by their aliases.

    :return: dict
    :rtype: dict
    """
    return _MESSAGES_REGISTRY


def get_registered_message_type(message_type):
    """Returns registered message type (class) by alias,

    :param str message_type: message type alias
    :return: MessageBase heirs instances.
    :rtype: MessageBase
    """
    try:
        return _MESSAGES_REGISTRY[message_type]
    except KeyError:
        raise UnknownMessageTypeError('`%s` message class is not registered' % message_type)


def import_app_sitemessage_module(app):
    """Returns a submodule of a given app

    :param str app: application name
    :return: submodule or None
    :rtype: module or None
    """
    module_name = settings.APP_MODULE_NAME
    module = import_module(app)
    try:
        sub_module = import_module('%s.%s' % (app, module_name))
        return sub_module
    except:
        if module_has_submodule(module, module_name):
            raise
        return None


def import_project_sitemessage_modules():
    """Imports sitemessages modules from registered apps."""
    from django.conf import settings as django_settings
    submodules = []
    for app in django_settings.INSTALLED_APPS:
        module = import_app_sitemessage_module(app)
        if module is not None:
            submodules.append(module)
    return submodules


def is_iterable(v):
    """Tells whether the thing is an iterable.
    NB: strings do not count even on Py3.

    """
    return hasattr(v, '__iter__') and not isinstance(v, six.string_types)


# Class used to represent message recipients.
Recipient = namedtuple('Recipient', ('messenger', 'user', 'address'))


class MessengerBase(object):
    """Base class for messengers used by sitemessage.

    Custom messenger classes, implementing various message delivery
    mechanics, other messenger classes must inherit from this one.

    """

    # Messenger alias to address it from different places, Should rather be quite unique %)
    alias = None
    _st = None  # Dispatches by status dict will be here. See init_delivery_statuses_dict().

    @classmethod
    def get_alias(cls):
        """Returns messenger alias.

        :return: str
        :rtype: str
        """
        if cls.alias is None:
            cls.alias = cls.__name__
        return cls.alias

    def __str__(self):
        return self.__class__.get_alias()

    def send_test_message(self, to, text):
        """Sends a test message using messengers settings.

        :param str to: an address to send test message to
        :param str text: text to send
        """
        self.before_send()
        result = self._test_message(to, text)
        self.after_send()
        return result

    def _test_message(self, to, text):
        """This method should be implemented by a heir to send a test message.
        
        :param str to: an address to send test message to
        :param str text: text to send
        """
        raise NotImplementedError(self.__class__.__name__ + ' must implement `test_message()`.')

    @classmethod
    def get_address(cls, recipient):
        """Returns recipient address.

        Heirs may override this to deduce address from `recipient` data
        (e.g. to get address from Django User model instance).

        :param object recipient: any object passed to `recipients()`
        :return: str
        :rtype: str
        """
        return recipient

    @classmethod
    def _structure_recipients_data(cls, recipients):
        """Converts recipients data into a list of Recipient objects.

        :param list recipients: list of objects
        :return: list of Recipient
        :rtype: list
        """
        if not is_iterable(recipients):
            recipients = (recipients,)

        objects = []
        for r in recipients:
            user = None
            if isinstance(r, USER_MODEL):
                user = r
            address = cls.get_address(r)  # todo maybe raise an exception of not a string?

            objects.append(Recipient(cls.get_alias(), user, address))

        return objects

    def _init_delivery_statuses_dict(self):
        """Initializes a dict indexed by message delivery statuses."""
        self._st = {
            'pending': [],
            'sent': [],
            'error': [],
            'failed': []
        }

    def mark_pending(self, dispatch):
        """Marks a dispatch as pending.

        Should be used within send().

        :param Dispatch dispatch: a Dispatch
        """
        self._st['pending'].append(dispatch)

    def mark_sent(self, dispatch):
        """Marks a dispatch as successfully sent.

        Should be used within send().

        :param Dispatch dispatch: a Dispatch
        """
        self._st['sent'].append(dispatch)

    def mark_error(self, dispatch, error_log, message_cls):
        """Marks a dispatch as having error or consequently as failed
        if send retry limit for that message type is exhausted.

        Should be used within send().

        :param Dispatch dispatch: a Dispatch
        :param str error_log: error message
        :param MessageBase message_cls: MessageBase heir
        """
        if message_cls.send_retry_limit is not None and (dispatch.retry_count + 1) >= message_cls.send_retry_limit:
            self.mark_failed(dispatch, error_log)
        else:
            dispatch.error_log = error_log
            self._st['error'].append(dispatch)

    def mark_failed(self, dispatch, error_log):
        """Marks a dispatch as failed.

        Sitemessage won't try to deliver already failed messages.

        Should be used within send().

        :param Dispatch dispatch: a Dispatch
        :param str error_log: str - error message
        """
        dispatch.error_log = error_log
        self._st['failed'].append(dispatch)

    def before_send(self):
        """This one is called right before send procedure.
        Usually heir will implement some messenger warm up (connect) code.

        """

    def after_send(self):
        """This one is called right after send procedure.
        Usually heir will implement some messenger cool down (disconnect) code.

        """

    def _process_messages(self, messages, ignore_unknown_message_types=False):
        """Performs message processing.

        :param dict messages: indexed by message id dict with messages data
        :param bool ignore_unknown_message_types: whether to silence exceptions
        :raises UnknownMessageTypeError:
        """
        self._init_delivery_statuses_dict()
        self.before_send()

        for message_id, message_data in messages.items():
            message_model, dispatch_models = message_data
            try:
                message_cls = get_registered_message_type(message_model.cls)
            except UnknownMessageTypeError:
                if ignore_unknown_message_types:
                    continue
                raise

            message_type_cache = None
            for dispatch in dispatch_models:
                if not dispatch.message_cache:  # Create actual message text for further usage.

                    if message_type_cache is None and not message_cls.has_dynamic_context:
                        # If a message class doesn't depend upon a dispatch data for message compilation, we'd compile a message just once.
                        message_type_cache = message_cls.compile(message_model, self)

                    dispatch.message_cache = message_type_cache or message_cls.compile(message_model, self, dispatch=dispatch)

            self.send(message_cls, message_model, dispatch_models)

        self.after_send()
        self._update_dispatches()

    def _update_dispatches(self):
        """Updates dispatched data in DB according to information gather by `mark_*` methods,"""
        Dispatch.log_dispatches_errors(self._st['error'] + self._st['failed'])
        Dispatch.set_dispatches_statuses(**self._st)
        self._init_delivery_statuses_dict()

    def send(self, message_cls, message_model, dispatch_models):
        """Main send method must be implement by all heirs.

        :param MessageBase message_cls: a MessageBase heir
        :param Message message_model: message model
        :param list dispatch_models: Dispatch models for this Message
        """
        raise NotImplementedError(self.__class__.__name__ + ' must implement `send()`.')


class MessageBase(object):
    """Base class for messages used by sitemessage.
    Customized message handling is available through inheritance.

    """

    # List of supported messengers (aliases).
    supported_messengers = []  #todo think over, implement

    # Message type alias to address it from different places, Should rather be quite unique %)
    alias = None

    # Number describing message priority. Can be overridden by `priority` provided with schedule_messages().
    priority = None

    # This flag is used to optimize template compilation process.
    # If True template will be compiled for every dispatch (and dispatch data will be available in it)
    # instead of just once per message.
    has_dynamic_context = False

    # Template file extension. Considered when the below mentioned `template` field is not set.
    template_ext = 'tpl'

    # Path to the template to be used for message rendering.
    # If not set, will be deduced from message, messenger data (e.g. `sitemessage/plain_smtp.txt`) and `template_ext` (see above).
    template = None

    # This limits the number of send attempts before message delivery considered failed.
    send_retry_limit = 10

    _message_model = None
    _dispatch_models = None

    SIMPLE_TEXT_ID = 'stext_'

    def __init__(self, context=None, template_path=None):
        """Initializes a message.

        :param dict, str context: data to be used for message rendering (e.g. in templates)
        :param str template_path: template path
        """
        context_base = {
            'tpl': None,  # Template path to use
            'use_tpl': False  # Use template boolean flag
        }

        if context is not None:
            self.update_context(context_base, context, template_path=template_path)

        self.context = context_base

    @classmethod
    def get_alias(cls):
        """Returns message type alias.

        :return: str
        :rtype: str
        """
        if cls.alias is None:
            cls.alias = cls.__name__
        return cls.alias

    def __str__(self):
        return self.__class__.get_alias()

    def get_context(self):
        """Returns message context.

        :return: dict
        :rtype: dict
        """
        return self.context

    def schedule(self, recipients=None, sender=None, priority=None):
        """Schedules message for a delivery.
        Puts message data into DB.

        :param list recipients: of Recipient
        :param User sender: Django User model heir instance
        :param int priority: number describing message priority
        :return: a tuple with message model and a list of dispatch models.
        :rtype: tuple
        """
        self._message_model, self._dispatch_models = Message.create(self.get_alias(), self.get_context(), recipients=recipients, sender=sender, priority=priority)
        return self._message_model, self._dispatch_models

    @classmethod
    def get_template(cls, message, messenger):
        """Get a template path to compile a message.

        1. `tpl` field of message context;
        2. `template` field of message class;
        3. deduced from message, messenger data and `template_ext` message type field (e.g. `sitemessage/plain_smtp.txt`).

        :param Message message: Message model
        :param MessengerBase messenger: a MessengerBase heir
        :return: str
        :rtype: str
        """
        template = message.context.get('tpl', None)

        if template:  # Template name is taken from message context.
            return template

        if cls.template is None:
            cls.template = 'sitemessage/%s_%s.%s' % (cls.get_alias(), messenger.get_alias(), cls.template_ext)
        return cls.template

    @classmethod
    def compile(cls, message, messenger, dispatch=None):
        """Compiles and returns a message text.

        Considers `use_tpl` field from message context to decide whether
        template compilation is used.

        Otherwise `text` field from message context is used as message contents.

        :param Message message: model instance
        :param MessengerBase messenger: MessengerBase heir instance
        :param Dispatch dispatch: model instance to consider context from
        :return: str
        :rtype: str
        """
        if message.context.get('use_tpl', False):
            context = message.context
            context.update({
                'message_model': message,
                'dispatch_model': dispatch
            })
            return render_to_string(cls.get_template(message, messenger), context)
        return message.context[cls.SIMPLE_TEXT_ID]

    @classmethod
    def update_context(cls, base_context, str_or_dict, template_path=None):
        """Helper method to structure message context data.

        NOTE: updates `base_context` inplace.

        :param dict base_context: context dict to update
        :param dict, str str_or_dict: text representing a message, or a dict to be placed into message context.
        :param str template_path: template path to be used for message rendering
        """
        if isinstance(str_or_dict, dict):
            base_context.update(str_or_dict)
            base_context['use_tpl'] = True
        else:
            base_context[cls.SIMPLE_TEXT_ID] = str_or_dict

        if cls.SIMPLE_TEXT_ID in str_or_dict:
            base_context['use_tpl'] = False

        base_context['tpl'] = template_path

    @classmethod
    def prepare_dispatches(cls, message):
        """Creates Dispatch models for a given message and return them.

        :param Message message: Message model instance
        :return: list of created Dispatch models
        :rtype: list
        """
        recipients = cls.calculate_recipients(message)
        return Dispatch.create(message, recipients)

    @classmethod
    def calculate_recipients(cls, message):
        """Calculates recipients for a given model.

        Method should be implemented by a heir to allow `prepare_dispatches()`.

        :param Message message: Message model instance
        :return: list of Recipient
        :rtype: list
        """
        raise NotImplementedError(cls.__name__ + ' must implement `calculate_recipients()` to use automatic recipient calculation used by `prepare_dispatches()`.')
